"""The `Installation` class is used to manage the `~/.{product}` folder on WorkspaceFS to track typed files."""

import csv
import dataclasses
import enum
import functools
import io
import json
import logging
import os.path
import re
import threading
import types
import typing
from collections.abc import Callable, Collection
from functools import partial
from json import JSONDecodeError
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Protocol,
    TypeVar,
    get_args,
    get_type_hints,
    runtime_checkable,
)

import databricks.sdk.core
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.mixins import workspace
from databricks.sdk.service.workspace import ImportFormat

from databricks.labs.blueprint.parallel import Threads

logger = logging.getLogger(__name__)

Json = dict[str, Any]

__all__ = ["Installation", "MockInstallation", "IllegalState", "NotInstalled", "SerdeError"]


class IllegalState(ValueError):
    pass


class NotInstalled(NotFound):
    """Raised when a product is not installed."""


class SerdeError(TypeError):
    """Raised when a serialization or deserialization error occurs."""


class Installation:
    """The `Installation` class is used to manage the `~/.{product}` folder on WorkspaceFS to track typed files.
    It provides methods for serializing and deserializing objects of a specific type, as well as managing the storage
    location for those objects. The class includes methods for loading and saving objects, uploading and downloading
    files, and managing the installation folder.

    The `Installation` class can be helpful for unit testing by allowing you to mock the file system and control
    the behavior of the `load` and `save` methods. You can create a `MockInstallation` object and use it to override
    the default installation folder and the contents of the files in that folder. This allows you to test the behavior
    of your code in different scenarios, such as when a file is not found or when the contents of a file do not match
    the expected format."""

    T = TypeVar("T")
    _PRIMITIVES = (int, bool, float, str)

    def __init__(self, ws: WorkspaceClient, product: str, *, install_folder: str | None = None):
        """The `Installation` class constructor creates an `Installation` object for the given product in
        the current workspace."""
        self._ws = ws
        self._product = product
        self._install_folder = install_folder
        self._lock = threading.Lock()

    @classmethod
    def current(cls, ws: WorkspaceClient, product: str, *, assume_user: bool = False) -> "Installation":
        """Returns the Installation object for the given product in the current workspace.

        If the installation is not found, a `NotInstalled` error is raised. If `assume_user` argument is True, the method
        will assume that the installation is in the user's home directory and return it if found. If False, the method
        will only return an installation that is in the `/Applications` directory."""
        user_folder = cls._user_home_installation(ws, product)
        applications_folder = cls._global_installation(product)
        folders = [user_folder, applications_folder]
        for candidate in folders:
            try:
                ws.workspace.get_status(candidate)
                return cls(ws, product, install_folder=candidate)
            except NotFound:
                logger.debug(f"{product} is not installed at {candidate}")
                continue
        if assume_user:
            return cls(ws, product, install_folder=user_folder)
        raise NotInstalled(f"Application not installed: {product}")

    @classmethod
    def assume_user_home(cls, ws: WorkspaceClient, product: str):
        """Constructs installation with explicit location in the user home folder"""
        user_folder = cls._user_home_installation(ws, product)
        return cls(ws, product, install_folder=user_folder)

    @classmethod
    def assume_global(cls, ws: WorkspaceClient, product: str):
        """Constructs installation with explicit global location in the /Applications folder"""
        applications_folder = cls._global_installation(product)
        return cls(ws, product, install_folder=applications_folder)

    @classmethod
    def existing(cls, ws: WorkspaceClient, product: str) -> Collection["Installation"]:
        """Returns a collection of all existing installations for the given product in the current workspace.

        This method searches for installations in the root /Applications directory and home directories of all users
        in the workspace."""

        def check_folder(install_folder: str) -> Installation | None:
            try:
                ws.workspace.get_status(install_folder)
                return cls(ws, product, install_folder=install_folder)
            except NotFound:
                return None

        tasks = [functools.partial(check_folder, cls._global_installation(product))]
        for user in ws.users.list(attributes="userName"):
            service_principal_folder = f"/Users/{user.user_name}/.{product}"
            tasks.append(functools.partial(check_folder, service_principal_folder))
        for service_principal in ws.service_principals.list(attributes="applicationId"):
            service_principal_folder = f"/Users/{service_principal.application_id}/.{product}"
            tasks.append(functools.partial(check_folder, service_principal_folder))
        return Threads.strict(f"finding {product} installations", tasks)

    @classmethod
    def load_local(cls, type_ref: type[T], file: Path) -> T:
        """Loads a typed file from the local file system."""
        with file.open("rb") as f:
            as_dict = cls._convert_content(file.name, f)
            return cls._unmarshal_type(as_dict, file.name, type_ref)

    def product(self) -> str:
        """The `product` method returns the name of the product associated with the installation."""
        return self._product

    def install_folder(self) -> str:
        """The `install_folder` method returns the path to the installation folder on WorkspaceFS.
        The installation folder is used to store typed files that are managed by the `Installation` class.

        If an `install_folder` argument is provided to the constructor of the `Installation` class, it will be used
        as the installation folder. Otherwise, the installation folder will be determined based on the current user's
        username. Specifically, the installation folder will be `/Users/{user_name}/.{product}`, where `{user_name}`
        is the username of the current user and `{product}` is the name of the product associated with the installation.

        Here is an example of how you can use the `install_folder` method:

        ```
        # Create an Installation object for the "blueprint" product
        install = Installation(WorkspaceClient(), "blueprint")

        # Print the path to the installation folder
        print(install.install_folder())
        # Output: /Users/{user_name}/.blueprint
        ```

        In this example, the `Installation` object is created for the "blueprint" product. The `install_folder` method
        is then called to print the path to the installation folder. The output will be `/Users/{user_name}/.blueprint`,
        where `{user_name}` is the username of the current user.

        You can also provide an `install_folder` argument to the constructor to specify a custom installation folder.
        Here is an example of how you can do this:

        ```
        # Create an Installation object for the "blueprint" product with a custom installation folder
        install = Installation(WorkspaceClient(), "blueprint", install_folder="/my/custom/folder")

        # Print the path to the installation folder
        print(install.install_folder())
        # Output: /my/custom/folder
        ```

        In this example, the `Installation` object is created for the "blueprint" product with a custom installation
        folder of `/my/custom/folder`. The `install_folder` method is then called to print the path to the installation
        folder. The output will be `/my/custom/folder`."""
        if self._install_folder is not None:
            return self._install_folder
        self._install_folder = self._user_home_installation(self._ws, self.product())
        return self._install_folder

    def is_global(self) -> bool:
        """Returns true if current installation is in /Applications folder"""
        return self.install_folder() == self._global_installation(self.product())

    def username(self) -> str:
        """Returns the username associated with the installation folder"""
        return os.path.basename(os.path.dirname(self.install_folder()))

    def load(self, type_ref: type[T], *, filename: str | None = None) -> T:
        """The `load` method loads an object of type `type_ref` from a file on WorkspaceFS. If no `filename` is
        provided, the `__file__` attribute of `type_ref` will be used as the filename.

        If the object has a `__version__` attribute, the method will check that the version of the object in the file
        matches the expected version. If the versions do not match, the method will attempt to migrate the object to
        the expected version using a method named `v{actual_version}_migrate` on the `type_ref` class. If the migration
        is successful, the method will return the migrated object. If the migration is not successful, the method will
        raise an `IllegalState` exception."""
        filename = self._get_filename(filename, type_ref)
        logger.debug(f"Loading {type_ref.__name__} from {filename}")
        as_dict = self._load_content(filename)
        return self._unmarshal_type(as_dict, filename, type_ref)

    def load_or_default(self, type_ref: type[T]) -> T:
        """If the file is not found, the method will return a default instance of the `type_ref` class."""
        try:
            return self.load(type_ref)
        except NotFound:
            filename = self._get_filename(None, type_ref)
            return self._unmarshal_type({}, filename, type_ref)

    def save(self, inst: T, *, filename: str | None = None):
        """The `save` method saves a dataclass object of type `T` to a file on WorkspaceFS.
        If no `filename` is provided, the name of the `type_ref` class will be used as the filename.
        If the object has a `__version__` attribute, the method will add a `version` field to the serialized object
        with the value of the `__version__` attribute.

        Here is an example of how you can use the `save` method:

        ```
        # Create an Installation object for the "blueprint" product
        install = Installation(WorkspaceClient(), "blueprint")

        # Save a dataclass object of type MyClass to a file
        @dataclasses.dataclass
        class MyClass:
            field1: str
            field2: str

        obj = MyClass(field1='value1', field2='value2')
        install.save(obj, filename='myfile.json')

        # Verify that the object was saved correctly
        loaded_obj = install.load(MyClass, filename='myfile.json')
        assert loaded_obj == obj
        ```

        In this example, the `Installation` object is created for the "blueprint" product. A dataclass object of type
        `MyClass` is then created and saved to a file using the `save` method. The object is then loaded from the file
        using the `load` method and compared to the original object to verify that it was saved correctly."""
        if not inst:
            raise SerdeError("missing value")
        type_ref = self._get_type_ref(inst)
        filename = self._get_filename(filename, type_ref)
        version = None
        if hasattr(inst, "__version__"):
            version = getattr(inst, "__version__")
        as_dict, _ = self._marshal(type_ref, [], inst)
        if version:
            as_dict["version"] = version
        self._overwrite_content(filename, as_dict, type_ref)
        return f"{self.install_folder()}/{filename}"

    def upload(self, filename: str, raw: bytes):
        """The `upload` method uploads raw bytes to a file on WorkspaceFS with the given `filename`. This method is
        used to upload files that are not typed, i.e., they do not have a corresponding `type_ref` object."""
        with self._lock:
            dst = f"{self.install_folder()}/{filename}"
            attempt = partial(self._ws.workspace.upload, dst, raw, format=ImportFormat.AUTO, overwrite=True)
            try:
                logger.debug(f"Uploading: {dst}")
                attempt()
            except NotFound as error:
                if error.error_code == "FEATURE_DISABLED":
                    self._enable_files_in_repos()
                parent_folder = os.path.dirname(dst)
                logger.debug(f"Creating missing folders: {parent_folder}")
                self._ws.workspace.mkdirs(parent_folder)
                attempt()
            return self._strip_notebook_source_suffix(dst, raw)

    _NOTEBOOK_MAGIC = {
        "py": b"# Databricks notebook source",
        "scala": b"// Databricks notebook source",
        "sql": b"-- Databricks notebook source",
    }

    @classmethod
    def _strip_notebook_source_suffix(cls, dst: str, raw: bytes) -> str:
        """If the file is a Databricks notebook, the method will remove the suffix from the filename."""
        if "." not in dst:
            return dst
        ext = dst.split(".")[-1]
        magic = cls._NOTEBOOK_MAGIC.get(ext)
        if not magic:
            return dst
        magic_len = len(magic)
        if len(raw) > magic_len and raw[0:magic_len] == magic:
            return dst.removesuffix(f".{ext}")
        return dst

    def upload_dbfs(self, filename: str, raw: BinaryIO) -> str:
        """The `upload_dbfs` method uploads raw bytes to a file on DBFS (Databricks File System) with the given
        `filename`. This method is used to upload files to DBFS, which is a distributed file system that is integrated
        with Databricks."""
        with self._lock:
            dst = f"{self.install_folder()}/{filename}"
            attempt = partial(self._ws.dbfs.upload, dst, raw, overwrite=True)
            try:
                logger.debug(f"Uploading to DBFS: {dst}")
                attempt()
            except NotFound:
                parent_folder = os.path.dirname(dst)
                logger.debug(f"Creating missing DBFS folders: {parent_folder}")
                self._ws.dbfs.mkdirs(parent_folder)
                attempt()
            return dst

    def files(self) -> list[workspace.ObjectInfo]:
        """The `files` method returns a list of all files in the installation folder on WorkspaceFS.
        This method is used to list the files that are managed by the `Installation` class."""
        return list(self._ws.workspace.list(self.install_folder(), recursive=True))

    def remove(self):
        """The `remove` method deletes the installation folder on WorkspaceFS.
        This method is used to remove all files and folders that are managed by the `Installation` class."""
        self._ws.workspace.delete(self.install_folder(), recursive=True)

    def workspace_link(self, path: str) -> str:
        """Returns a link to a file in a workspace.

        Usage:
            >>> import webbrowser
            >>> installation = Installation.current()
            >>> webbrowser.open(installation.workspace_link('config.yml'))
        """
        return f"{self._host()}/#workspace{self.install_folder()}/{path.removeprefix('/')}"

    def workspace_markdown_link(self, label: str, path: str) -> str:
        """Returns a markdown link to a file in a workspace."""
        return f"[{label}]({self.workspace_link(path)})"

    def _host(self):
        """Returns the host of the current workspace."""
        return self._ws.config.host

    def _overwrite_content(self, filename: str, as_dict: Json, type_ref: type):
        """The `_overwrite_content` method is a private method that is used to serialize an object of type `type_ref`
        to a dictionary and write it to a file on WorkspaceFS. This method is called by the `save` and `upload` methods.

        The `as_dict` argument is the dictionary representation of the object that is to be written to the file.
        The `type_ref` argument is the type of the object that is being saved."""
        converters: dict[str, Callable[[Any, type], bytes]] = {
            "json": self._dump_json,
            "yml": self._dump_yaml,
            "csv": self._dump_csv,
        }
        extension = filename.split(".")[-1]
        if extension not in converters:
            raise KeyError(f"Unknown extension: {extension}")
        logger.debug(f"Converting {type_ref.__name__} into {extension.upper()} format")
        raw = converters[extension](as_dict, type_ref)
        self.upload(filename, raw)

    @staticmethod
    def _global_installation(product):
        """The `_global_installation` method is a private method that is used to determine the installation folder
        for the given product in the `/Applications` directory. This method is called by the `install_folder` method."""
        return f"/Applications/{product}"

    @classmethod
    def _unmarshal_type(cls, as_dict, filename, type_ref):
        """The `_unmarshal_type` method is a private method that is used to deserialize a dictionary to an object of
        type `type_ref`. This method is called by the `load` method."""
        expected_version = None
        if hasattr(type_ref, "__version__"):
            expected_version = getattr(type_ref, "__version__")
        if expected_version:
            as_dict = cls._migrate_file_format(type_ref, expected_version, as_dict, filename)
        return cls._unmarshal(as_dict, [], type_ref)

    def _load_content(self, filename: str) -> Json:
        """The `_load_content` method is a private method that is used to load the contents of a file from
        WorkspaceFS as a dictionary. This method is called by the `load` method."""
        with self._lock:
            # TODO: check how to make this fail fast during unit testing, otherwise
            # this currently hangs with the real installation class and mocked workspace client
            with self._ws.workspace.download(f"{self.install_folder()}/{filename}") as f:
                return self._convert_content(filename, f)

    @classmethod
    def _convert_content(cls, filename: str, raw: BinaryIO) -> Json:
        """The `_convert_content` method is a private method that is used to convert the raw bytes of a file to a
        dictionary. This method is called by the `_load_content` method."""
        converters: dict[str, Callable[[BinaryIO], Any]] = {
            "json": json.load,
            "yml": cls._load_yaml,
            "csv": cls._load_csv,
        }
        extension = filename.split(".")[-1]
        if extension not in converters:
            raise KeyError(f"Unknown extension: {extension}")
        try:
            return converters[extension](raw)
        except JSONDecodeError:
            return {}

    def __repr__(self):
        return self.install_folder()

    def __eq__(self, other):
        if not isinstance(other, Installation):
            return False
        return self.install_folder() == other.install_folder()

    def __hash__(self):
        """The `__hash__` method is used to hash the `Installation` object.
        This method is called by the `hash` function."""
        return hash(self.install_folder())

    @staticmethod
    def _user_home_installation(ws: WorkspaceClient, product: str):
        """The `_user_home_installation` method is a private method that is used to determine the installation folder
        for the current user. This method is called by the `install_folder` method."""
        me = ws.current_user.me()
        return f"/Users/{me.user_name}/.{product}"

    @staticmethod
    def _migrate_file_format(type_ref, expected_version, as_dict, filename):
        """The `_migrate_file_format` method is a private method that is used to migrate the file format of a file"""
        actual_version = as_dict.pop("version", 1)
        while actual_version < expected_version:
            migrate = getattr(type_ref, f"v{actual_version}_migrate", None)
            if not migrate:
                break
            as_dict = migrate(as_dict)
            prev_version = actual_version
            actual_version = as_dict.pop("version", 1)
            if actual_version == prev_version:
                raise IllegalState(f"cannot migrate {filename} from v{prev_version}")
        if actual_version != expected_version:
            raise IllegalState(f"expected state version={expected_version}, got={actual_version}")
        return as_dict

    @staticmethod
    def _get_filename(filename: str | None, type_ref: type) -> str:
        """The `_get_filename` method is a private method that is used to determine the filename of a file based on
        the type of the object that is being saved. This method is called by the `save` method."""
        if not filename and hasattr(type_ref, "__file__"):
            return getattr(type_ref, "__file__")
        if not filename:
            kebab_name = re.sub(r"(?<!^)(?=[A-Z])", "-", type_ref.__name__).lower()
            filename = f"{kebab_name}.json"
        return filename

    @classmethod
    def _get_type_ref(cls, inst) -> type:
        """The `_get_type_ref` method is a private method that is used to determine the type of an object. This method
        is called by the `save` method."""
        type_ref = type(inst)
        if type_ref == list:
            return cls._get_list_type_ref(inst)
        return type_ref

    @staticmethod
    def _get_list_type_ref(inst: T) -> type[list[T]]:
        """The `_get_list_type_ref` method is a private method that is used to determine the type of a list object."""
        from_list: list = inst  # type: ignore[assignment]
        if len(from_list) == 0:
            raise ValueError("List cannot be empty")
        item_type = type(from_list[0])  # type: ignore[misc]
        return list[item_type]  # type: ignore[valid-type]

    def _marshal(self, type_ref: type, path: list[str], inst: Any) -> tuple[Any, bool]:
        """The `_marshal` method is a private method that is used to serialize an object of type `type_ref` to
        a dictionary. This method is called by the `save` method."""
        if inst is None:
            return None, False
        if isinstance(inst, databricks.sdk.core.Config):
            return self._marshal_databricks_config(inst)
        if hasattr(inst, "as_dict"):
            return inst.as_dict(), True
        if dataclasses.is_dataclass(type_ref):
            return self._marshal_dataclass(type_ref, path, inst)
        if type_ref == list:
            return self._marshal_list(type_ref, path, inst)
        if isinstance(type_ref, enum.EnumMeta):
            return self._marshal_enum(inst)
        if type_ref == types.NoneType:
            return inst, inst is None
        if type_ref in self._PRIMITIVES:
            return inst, True
        return self._marshal_generic_types(type_ref, path, inst)

    def _marshal_generic_types(self, type_ref: type, path: list[str], inst: Any) -> tuple[Any, bool]:
        # pylint: disable-next=import-outside-toplevel,import-private-name
        from typing import (  # type: ignore[attr-defined]
            _GenericAlias,
            _UnionGenericAlias,
        )

        if isinstance(type_ref, (types.UnionType, _UnionGenericAlias)):  # type: ignore[attr-defined]
            return self._marshal_union(type_ref, path, inst)
        if isinstance(type_ref, (_GenericAlias, types.GenericAlias)):  # type: ignore[attr-defined]
            if type_ref.__origin__ in (dict, list) or isinstance(type_ref, types.GenericAlias):
                return self._marshal_generic(type_ref, path, inst)
            return self._marshal_generic_alias(type_ref, inst)
        raise SerdeError(f'{".".join(path)}: unknown: {inst}')

    def _marshal_union(self, type_ref: type, path: list[str], inst: Any) -> tuple[Any, bool]:
        """The `_marshal_union` method is a private method that is used to serialize an object of type `type_ref` to
        a dictionary. This method is called by the `save` method."""
        combo = []
        for variant in get_args(type_ref):
            value, ok = self._marshal(variant, [*path, f"(as {variant})"], inst)
            if ok:
                return value, True
            combo.append(self._explain_why(variant, [*path, f"(as {variant})"], inst))
        raise SerdeError(f'{".".join(path)}: union: {" or ".join(combo)}')

    def _marshal_generic(self, type_ref: type, path: list[str], inst: Any) -> tuple[Any, bool]:
        """The `_marshal_generic` method is a private method that is used to serialize an object of type `type_ref`
        to a dictionary. This method is called by the `save` method."""
        type_args = get_args(type_ref)
        if not type_args:
            raise SerdeError(f"Missing type arguments: {type_args}")
        if len(type_args) == 2:
            return self._marshal_dict(type_args[1], path, inst)
        return self._marshal_list(type_args[0], path, inst)

    @staticmethod
    def _marshal_generic_alias(type_ref, inst):
        """The `_marshal_generic_alias` method is a private method that is used to serialize an object of type
        `type_ref` to a dictionary. This method is called by the `save` method."""
        if not inst:
            return None, False
        return inst, isinstance(inst, type_ref.__origin__)  # type: ignore[attr-defined]

    def _marshal_list(self, type_ref: type, path: list[str], inst: Any) -> tuple[Any, bool]:
        """The `_marshal_list` method is a private method that is used to serialize an object of type `type_ref` to
        a dictionary. This method is called by the `save` method."""
        as_list = []
        if not isinstance(inst, list):
            return None, False
        for i, v in enumerate(inst):
            value, ok = self._marshal(type_ref, [*path, f"{i}"], v)
            if not ok:
                raise SerdeError(self._explain_why(type_ref, [*path, f"{i}"], v))
            as_list.append(value)
        return as_list, True

    def _marshal_dict(self, type_ref: type, path: list[str], inst: Any) -> tuple[Any, bool]:
        """The `_marshal_dict` method is a private method that is used to serialize an object of type `type_ref` to
        a dictionary. This method is called by the `save` method."""
        if not isinstance(inst, dict):
            return None, False
        as_dict = {}
        for k, v in inst.items():
            as_dict[k], ok = self._marshal(type_ref, [*path, k], v)
            if not ok:
                raise SerdeError(self._explain_why(type_ref, [*path, k], v))
        return as_dict, True

    def _marshal_dataclass(self, type_ref: type, path: list[str], inst: Any) -> tuple[Any, bool]:
        """The `_marshal_dataclass` method is a private method that is used to serialize an object of type `type_ref`
        to a dictionary. This method is called by the `save` method."""
        if inst is None:
            return None, False
        as_dict = {}
        for field, hint in get_type_hints(type_ref).items():
            origin = getattr(hint, "__origin__", None)
            if origin is typing.ClassVar:
                continue
            raw = getattr(inst, field)
            if raw is None:
                continue
            value, ok = self._marshal(hint, [*path, field], raw)
            if not ok:
                raise SerdeError(self._explain_why(hint, [*path, field], raw))
            if not value:
                continue
            as_dict[field] = value
        return as_dict, True

    def _marshal_databricks_config(self, inst):
        """The `_marshal_databricks_config` method is a private method that is used to serialize an object of type
        `databricks.sdk.core.Config` to a dictionary. This method is called by the `save` method."""
        if not inst:
            return None, False
        current_client_config = self._current_client_config()
        remote_file_config = inst.as_dict()
        if current_client_config == remote_file_config:
            return None, True
        return remote_file_config, True

    def _current_client_config(self) -> dict:
        return self._ws.config.as_dict()

    @staticmethod
    def _marshal_enum(inst):
        """The `_marshal_enum` method is a private method that is used to serialize an object of type `enum.Enum` to
        a dictionary. This method is called by the `save` method."""
        if not inst:
            return None, False
        return inst.value, True

    @runtime_checkable
    class _FromDict(Protocol):
        """The `_FromDict` protocol is used to define a type that can be constructed from a dictionary. This protocol
        is used to define a type that can be constructed from a dictionary. This protocol is used to define a type that
        can be constructed from a dictionary."""

        @classmethod
        def from_dict(cls, raw: dict):
            pass

    @classmethod
    def _unmarshal(cls, inst: Any, path: list[str], type_ref: type[T]) -> T | None:
        """The `_unmarshal` method is a private method that is used to deserialize a dictionary to an object of type
        `type_ref`. This method is called by the `load` method."""
        if dataclasses.is_dataclass(type_ref):
            return cls._unmarshal_dataclass(inst, path, type_ref)
        if isinstance(type_ref, enum.EnumMeta):
            if not inst:
                return None
            return type_ref(inst)
        if type_ref in cls._PRIMITIVES:
            return cls._unmarshal_primitive(inst, type_ref)
        if type_ref == databricks.sdk.core.Config:
            if not inst:
                inst = {}
            return databricks.sdk.core.Config(**inst)  # type: ignore[return-value]
        if type_ref == types.NoneType:
            return None
        if isinstance(type_ref, cls._FromDict):
            return type_ref.from_dict(inst)
        return cls._unmarshal_generic_types(type_ref, path, inst)

    @classmethod
    def _unmarshal_generic_types(cls, type_ref, path, inst):
        # pylint: disable-next=import-outside-toplevel,import-private-name
        from typing import (  # type: ignore[attr-defined]
            _GenericAlias,
            _UnionGenericAlias,
        )

        if isinstance(type_ref, (types.UnionType, _UnionGenericAlias)):
            return cls._unmarshal_union(inst, path, type_ref)
        if isinstance(type_ref, (_GenericAlias, types.GenericAlias)):
            return cls._unmarshal_generic(inst, path, type_ref)
        raise SerdeError(f'{".".join(path)}: unknown: {type_ref}: {inst}')

    @classmethod
    def _unmarshal_dataclass(cls, inst, path, type_ref):
        """The `_unmarshal_dataclass` method is a private method that is used to deserialize a dictionary to an object
        of type `type_ref`. This method is called by the `load` method."""
        if inst is None:
            return None
        if not isinstance(inst, dict):
            raise SerdeError(cls._explain_why(dict, path, inst))
        from_dict = {}
        fields = getattr(type_ref, "__dataclass_fields__")
        for field_name, hint in get_type_hints(type_ref).items():
            origin = getattr(hint, "__origin__", None)
            if origin is typing.ClassVar:
                continue
            raw = inst.get(field_name)
            value = cls._unmarshal(raw, [*path, field_name], hint)
            if value is None:
                field = fields.get(field_name)
                default_value = field.default
                default_factory = field.default_factory
                if default_factory == dataclasses.MISSING and default_value == dataclasses.MISSING:
                    raise SerdeError(cls._explain_why(hint, [*path, field_name], value))
                if default_value != dataclasses.MISSING:
                    value = default_value
                else:
                    value = default_factory()
            from_dict[field_name] = value
        return type_ref(**from_dict)

    @classmethod
    def _unmarshal_union(cls, inst, path, type_ref):
        """The `_unmarshal_union` method is a private method that is used to deserialize a dictionary to an object
        of type `type_ref`. This method is called by the `load` method."""
        for variant in get_args(type_ref):
            value = cls._unmarshal(inst, path, variant)
            if value:
                return value
        return None

    @classmethod
    def _unmarshal_generic(cls, inst, path, type_ref):
        """The `_unmarshal_generic` method is a private method that is used to deserialize a dictionary to an object
        of type `type_ref`. This method is called by the `load` method."""
        # pylint: disable-next=import-outside-toplevel,import-private-name
        from typing import _GenericAlias  # type: ignore[attr-defined]

        type_args = get_args(type_ref)
        if not type_args:
            raise SerdeError(f"Missing type arguments: {type_args}")
        if type_ref.__origin__ not in (dict, list) and isinstance(type_ref, _GenericAlias):
            return cls._unmarshal(inst, path, type_ref.__origin__)
        if inst is None:
            return None
        if len(type_args) == 2:
            return cls._unmarshal_dict(inst, path, type_args[1])
        return cls._unmarshal_list(inst, path, type_args[0])

    @classmethod
    def _unmarshal_list(cls, inst, path, hint):
        """The `_unmarshal_list` method is a private method that is used to deserialize a dictionary to an object
        of type `type_ref`. This method is called by the `load` method."""
        if inst is None:
            return None
        as_list = []
        for i, v in enumerate(inst):
            as_list.append(cls._unmarshal(v, [*path, f"{i}"], hint))
        return as_list

    @classmethod
    def _unmarshal_dict(cls, inst, path, type_ref):
        """The `_unmarshal_dict` method is a private method that is used to deserialize a dictionary to an object
        of type `type_ref`. This method is called by the `load` method."""
        if inst is None:
            return None
        if not isinstance(inst, dict):
            raise SerdeError(cls._explain_why(type_ref, path, inst))
        from_dict = {}
        for k, v in inst.items():
            from_dict[k] = cls._unmarshal(v, [*path, k], type_ref)
        return from_dict

    @classmethod
    def _unmarshal_primitive(cls, inst, type_ref):
        """The `_unmarshal_primitive` method is a private method that is used to deserialize a dictionary to an object
        of type `type_ref`. This method is called by the `load` method."""
        if not inst:
            return inst
        # convert from str to int if necessary
        converted = type_ref(inst)  # type: ignore[call-arg]
        return converted

    @staticmethod
    def _explain_why(type_ref: type, path: list[str], raw: Any) -> str:
        """The `_explain_why` method is a private method that is used to explain why a value is not of the expected
        type. This method is called by the `_unmarshal` and `_marshal` methods."""
        if raw is None:
            raw = "value is missing"
        return f'{".".join(path)}: not a {type_ref.__name__}: {raw}'

    @staticmethod
    def _dump_json(as_dict: Json, _: type) -> bytes:
        """The `_dump_json` method is a private method that is used to serialize a dictionary to a JSON string. This
        method is called by the `save` method."""
        return json.dumps(as_dict, indent=2).encode("utf8")

    @staticmethod
    def _dump_yaml(raw: Json, _: type) -> bytes:
        """The `_dump_yaml` method is a private method that is used to serialize a dictionary to a YAML string. This
        method is called by the `save` method."""
        try:
            from yaml import dump  # pylint: disable=import-outside-toplevel

            return dump(raw).encode("utf8")
        except ImportError as err:
            raise SyntaxError("PyYAML is not installed. Fix: pip install databricks-labs-blueprint[yaml]") from err

    @staticmethod
    def _load_yaml(raw: BinaryIO) -> Json:
        """The `_load_yaml` method is a private method that is used to deserialize a YAML string to a dictionary. This
        method is called by the `load` method."""
        try:
            from yaml import (  # pylint: disable=import-outside-toplevel
                YAMLError,
                safe_load,
            )

            try:
                return safe_load(raw)
            except YAMLError as err:
                raise JSONDecodeError(str(err), "<yaml>", 0) from err
        except ImportError as err:
            raise SyntaxError("PyYAML is not installed. Fix: pip install databricks-labs-blueprint[yaml]") from err

    @staticmethod
    def _dump_csv(raw: list[Json], type_ref: type) -> bytes:
        """The `_dump_csv` method is a private method that is used to serialize a list of dictionaries to a CSV string.
        This method is called by the `save` method."""
        type_args = get_args(type_ref)
        if not type_args:
            raise SerdeError(f"Writing CSV is only supported for lists. Got {type_ref}")
        dataclass_ref = type_args[0]
        if not dataclasses.is_dataclass(dataclass_ref):
            raise SerdeError(f"Only lists of dataclasses allowed. Got {dataclass_ref}")
        non_empty_keys = set()
        for as_dict in raw:
            if not isinstance(as_dict, dict):
                raise SerdeError(f"Expecting a list of dictionaries. Got {as_dict}")
            for k, v in as_dict.items():
                if not v:
                    continue
                non_empty_keys.add(k)
        buffer = io.StringIO()
        # get ordered field names the way they appear in dataclass
        field_names = [_.name for _ in dataclasses.fields(dataclass_ref) if _.name in non_empty_keys]
        writer = csv.DictWriter(buffer, field_names, dialect="excel")
        writer.writeheader()
        for as_dict in raw:
            writer.writerow(as_dict)
        buffer.seek(0)
        return buffer.read().encode("utf8")

    @staticmethod
    def _load_csv(raw: BinaryIO) -> list[Json]:
        with io.TextIOWrapper(raw, encoding="utf8") as text_file:
            out = []
            for row in csv.DictReader(text_file):  # type: ignore[arg-type]
                out.append(row)
            return out

    def _enable_files_in_repos(self):
        """The `_enable_files_in_repos` method is a private method that is used to enable the "Files In Repos"
        feature on the current workspace. This method is called by the `upload` method."""
        workspace_file_system = self._ws.workspace_conf.get_status("enableWorkspaceFilesystem")

        logger.debug("Checking Files In Repos configuration")

        if workspace_file_system["enableWorkspaceFilesystem"] == "false":
            logger.debug("enableWorkspaceFilesystem is False, enabling the config")
            self._ws.workspace_conf.set_status({"enableWorkspaceFilesystem": "true"})


class MockInstallation(Installation):
    """Install state testing toolbelt

    register with PyTest:

        pytest.register_assert_rewrite('databricks.labs.blueprint.installation')
    """

    def __init__(self, overwrites: Any = None, *, is_global=True):  # pylint: disable=super-init-not-called
        if not overwrites:
            overwrites = {}
        self._overwrites = overwrites
        self._uploads: dict[str, bytes] = {}
        self._dbfs: dict[str, bytes] = {}
        self._removed = False
        self._is_global = is_global

    def install_folder(self) -> str:
        return "~/mock"

    def is_global(self) -> bool:
        return self._is_global

    def product(self) -> str:
        return "mock"

    def _host(self):
        return "https://localhost"

    def upload(self, filename: str, raw: bytes):
        self._uploads[filename] = raw
        return f"{self.install_folder()}/{filename}"

    def upload_dbfs(self, filename: str, raw: BinaryIO) -> str:
        self._dbfs[filename] = raw.read()
        return f"{self.install_folder()}/{filename}"

    def files(self) -> list[workspace.ObjectInfo]:
        out = []
        for filename in self._overwrites.keys():
            out.append(
                workspace.ObjectInfo(
                    path=f"{self.install_folder()}/{filename}",
                    object_type=workspace.ObjectType.FILE,
                )
            )
        for filename in self._uploads:
            out.append(
                workspace.ObjectInfo(
                    path=f"{self.install_folder()}/{filename}",
                    object_type=workspace.ObjectType.FILE,
                )
            )
        return out

    def remove(self):
        self._removed = True

    def _current_client_config(self) -> dict:
        return {}

    def _overwrite_content(self, filename: str, as_dict: Json, type_ref: type):
        self._overwrites[filename] = as_dict

    def _load_content(self, filename: str) -> Json:
        if filename not in self._overwrites:
            raise NotFound(filename)
        return self._overwrites[filename]

    def assert_file_written(self, filename: str, expected: Any):
        assert filename in self._overwrites, f"{filename} had no writes"
        if isinstance(expected, dict):
            for k, v in expected.items():
                if v == ...:
                    self._overwrites[filename][k] = ...
        actual = self._overwrites[filename]
        assert expected == actual, f"{filename} content missmatch"

    def assert_file_uploaded(self, filename, expected: bytes | None = None):
        """Asserts that a file was uploaded with the expected content"""
        self._assert_upload(filename, self._uploads, expected)

    def assert_file_dbfs_uploaded(self, filename, expected: bytes | None = None):
        """Asserts that a file was uploaded to DBFS with the expected content"""
        self._assert_upload(filename, self._dbfs, expected)

    def assert_removed(self):
        assert self._removed

    @staticmethod
    def _assert_upload(filename: Any, loc: dict[str, bytes], expected: bytes | None = None):
        if isinstance(filename, re.Pattern):
            for name in loc.keys():
                if not filename.match(name):
                    continue
                if expected:
                    assert loc[name] == expected, f"{filename} content missmatch"
                return
            raise AssertionError(f'Cannot find {filename.pattern} among {", ".join(loc.keys())}')
        assert filename in loc, f"{filename} had no writes"
        if expected:
            assert loc[filename] == expected, f"{filename} content missmatch"



"""Databricks' path specification, in the same vein as :module:`posixpath` and :module:`ntpath`.

Paths are Posix-like, but we don't use the builtin :module:`posixpath` directly because as of Python 3.13 the module
itself is part of a path's identity for the purposes of comparison.
"""

from posixpath import join, sep

__all__ = ("join", "sep")



"""Automated rollout of application upgrades deployed in a Databricks workspace."""

import importlib.util
import logging
from dataclasses import dataclass, field
from pathlib import Path

from databricks.sdk import WorkspaceClient
from databricks.sdk.mixins.compute import SemVer

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.wheels import ProductInfo, Version

logger = logging.getLogger(__name__)


@dataclass
class AppliedUpgrades:
    __version__ = 1
    upgrades: list[str] = field(default_factory=list)


class Upgrades:
    """Automated rollout of application upgrades deployed in a Databricks workspace.

    As time goes by, your applications evolve as well, requiring the addition of new columns to database schemas,
    changes of the database state, or some migrations of configured workflows. This utility allows you to do seamless
    upgrades from version X to version Z through version Y. Idiomatic usage in your deployment automation is as follows:

    >>> from ... import Config
    >>> from databricks.sdk import WorkspaceClient
    >>> from databricks.labs.blueprint.upgrades import Upgrades
    >>> from databricks.labs.blueprint.wheels import ProductInfo
    >>>
    >>> product_info = ProductInfo.from_class(Config)
    >>> ws = WorkspaceClient(product=product_info.product_name(), product_version=product_info.version())
    >>> installation = product_info.current_installation(ws)
    >>> config = installation.load(Config)
    >>> upgrades = Upgrades(product_info, installation)
    >>> upgrades.apply(ws)

    The upgrade process loads the version of the product that is about to be installed from `__about__.py` file that
    declares the `__version__` variable. This version is compares with the version currently installed on
    the Databricks Workspace by loading it from the `version.json` file in the installation folder. This file is kept
    up-to-date automatically if you use the databricks.labs.blueprint.wheels.WheelsV2.

    If those versions are different, the process looks for the `upgrades` folder next to `__about__.py` file and
    computes a difference for the upgrades in need to be rolled out. Every upgrade script in that directory has to
    start with a valid SemVer identifier, followed by the alphanumeric description of the change,
    like `v0.0.1_add_service.py`. Each script has to expose a function that takes `Installation` and
    `WorkspaceClient` arguments to perform the relevant upgrades. Here's the example:

    >>> import logging, dataclasses
    >>> from databricks.sdk import WorkspaceClient
    >>> from databricks.labs.blueprint.installation import Installation
    >>> from ... import Config
    >>> upgrade_logger = logging.getLogger(__name__)
    >>>
    >>> def upgrade(installation: Installation, ws: WorkspaceClient):
    >>>     upgrade_logger.info(f"creating new automated service user for the installation")
    >>>     config = installation.load(Config)
    >>>     service_principal = ws.service_principals.create(display_name='blueprint-service')
    >>>     new_config = dataclasses.replace(config, application_id=service_principal.application_id)
    >>>     installation.save(new_config)

    To prevent the same upgrade script from being applies twice, we use `applied-upgrades.json` file in
    the installation directory.
    """

    def __init__(self, product_info: ProductInfo, installation: Installation):
        self._product_info = product_info
        self._installation = installation

    def apply(self, ws: WorkspaceClient):
        """Applies application state upgrades for the given installation."""
        upgrades_folder = self._product_info.version_file().parent / "upgrades"
        if not upgrades_folder.exists():
            logger.warning(f"No upgrades folder: {upgrades_folder}")
            return
        applied = self._installation.load_or_default(AppliedUpgrades)
        for script in sorted(self._diff(upgrades_folder)):
            if script.name in applied.upgrades:
                logger.info(f"Already applied: {script.name}")
                continue
            self._apply_python_script(script, ws)
            applied.upgrades.append(script.name)
            self._installation.save(applied)

    def _apply_python_script(self, script: Path, ws: WorkspaceClient):
        """Load and apply the upgrade script."""
        name = "_".join(script.name.removesuffix(".py").split("_")[1:])
        spec = importlib.util.spec_from_file_location(name, script.as_posix())
        if not spec:
            logger.warning(f"Cannot load: {script.name}")
            return
        change = importlib.util.module_from_spec(spec)
        if not spec.loader:
            logger.warning(f"No loader: {script.name}")
            return
        spec.loader.exec_module(change)
        if not hasattr(change, "upgrade"):
            logger.warning(f"No upgrade(installation, ws) callback: {script.name}")
            return
        logger.info(f"Applying {script.name}...")
        change.upgrade(self._installation, ws)

    def _installed(self) -> SemVer:
        """Load the installed version of the product."""
        return self._installation.load(Version).as_semver()

    def _diff(self, upgrades_folder: Path):
        """Yield the upgrade scripts that need to be applied."""
        current = self._product_info.as_semver()
        installed_version = self._installed()
        for file in upgrades_folder.glob("v*.py"):
            try:
                semver = self._parse_version(file.name)
            except ValueError:
                logger.warning(f"not an upgrade script: {file.name}")
                continue
            if semver < installed_version:
                continue
            if semver > current:
                logger.warning(f"future version: {file.name}")
                continue
            yield file

    @staticmethod
    def _parse_version(name: str) -> SemVer:
        """Parse the version from the upgrade script name."""
        split = name.split("_")
        if len(split) < 2:
            raise ValueError(f"invalid spec: {name}")
        return SemVer.parse(split[0])



import os

from databricks.sdk.core import with_user_agent_extra

from databricks.labs.blueprint.__about__ import __version__

with_user_agent_extra("blueprint", __version__)

cli_version = os.environ.get("DATABRICKS_CLI_VERSION")
if cli_version:
    with_user_agent_extra("cli", cli_version)



"""A nice formatter for logging. It uses colors and bold text if the console supports it."""

import logging
import sys
from typing import TextIO


class NiceFormatter(logging.Formatter):
    """A nice formatter for logging. It uses colors and bold text if the console supports it."""

    # TODO: Actually detect if the console supports colors. Currently, it just assumes that it does.

    BOLD = "\033[1m"
    RESET = "\033[0m"
    GREEN = "\033[32m"
    BLACK = "\033[30m"
    CYAN = "\033[36m"
    YELLOW = "\033[33m"
    RED = "\033[31m"
    MAGENTA = "\033[35m"
    GRAY = "\033[90m"

    colors: bool
    """Whether this formatter is formatting with colors or not."""

    _levels: dict[int, str]
    """The colorized level names for each logging level."""

    _msg_colors: dict[int, str]
    """The color codes to use for rendering the message text depending on the logging level."""

    def __init__(self, *, probe_tty: bool = False, stream: TextIO = sys.stdout) -> None:
        """Create a new instance of the formatter.

        Args:
            stream: the output stream to which the formatter will write, used to check if it is a console.
            probe_tty: If true, the formatter will enable color support if the output stream appears to be a console.
        """
        super().__init__(fmt="%(asctime)s %(levelname)s [%(name)s] %(message)s", datefmt="%H:%M:%S")
        # Used to colorize the level names.
        self._levels = {
            logging.DEBUG: self._bold(f"{self.CYAN}   DEBUG"),
            logging.INFO: self._bold(f"{self.GREEN}    INFO"),
            logging.WARNING: self._bold(f"{self.YELLOW} WARNING"),
            logging.ERROR: self._bold(f"{self.RED}   ERROR"),
            logging.CRITICAL: self._bold(f"{self.MAGENTA}CRITICAL"),
        }
        # Used to colorize the message text. (These are prefixes: after the message text, the color is reset.)
        self._msg_colors = {
            logging.DEBUG: self.GRAY,
            logging.INFO: self.BOLD,
            logging.WARNING: self.BOLD,
            logging.ERROR: f"{self.BOLD}{self.RED}",
            logging.CRITICAL: f"{self.BOLD}{self.RED}",
        }
        # show colors in runtime, github actions, and while debugging
        self.colors = stream.isatty() if probe_tty else True

    def _bold(self, text: str) -> str:
        """Return text in bold."""
        return f"{self.BOLD}{text}{self.RESET}"

    def format(self, record: logging.LogRecord) -> str:
        """Format the log record. If colors are enabled, use them."""
        if not self.colors:
            return super().format(record)
        timestamp = self.formatTime(record, datefmt="%H:%M:%S")
        level = self._levels[record.levelno]

        # databricks.labs.ucx.foo.bar -> d.l.u.foo.bar
        module_split = record.name.split(".")
        abbreviated = [c[:1] for c in module_split[:-2]]  # abbreviate all but the last two components
        as_is = module_split[-2:]  # keep the last two components as-is
        name = ".".join([*abbreviated, *as_is])

        msg = record.getMessage()
        if record.exc_info and not record.exc_text:
            record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if not msg.endswith("\n"):
                msg += "\n"
            msg += record.exc_text
        if record.stack_info:
            if not msg.endswith("\n"):
                msg += "\n"
            msg += self.formatStack(record.stack_info)

        color_marker = self._msg_colors[record.levelno]

        thread_name = f"[{record.threadName}]" if record.threadName != "MainThread" else ""
        return f"{self.GRAY}{timestamp}{self.RESET} {level} {color_marker}[{name}]{thread_name} {msg}{self.RESET}"


def install_logger(
    level: int | str = logging.DEBUG, *, stream: TextIO = sys.stderr, root: logging.Logger = logging.root
) -> logging.StreamHandler:
    """Install a console logger with a nice formatter.

    The root logger will be modified:

     - Its logging level will be left as-is.
     - All existing handlers will be removed.
     - A new handler will be installed with our custom formatter. It will be configured to emit logs at the given level
       (default: DEBUG) or higher, to the specified stream (default: sys.stderr).

    Args:
        level: The logging level to set for the console handler.
        stream: The stream to which the logger will write. Defaults to sys.stderr.
        root: The root logger to modify. Defaults to the system root logger. (Mainly useful in tests.)

    Returns:
        The logging handler that was installed.
    """
    for handler in root.handlers:
        root.removeHandler(handler)
    console_handler = logging.StreamHandler(stream)
    console_handler.setFormatter(NiceFormatter(stream=stream))
    console_handler.setLevel(level)
    root.addHandler(console_handler)
    return console_handler



"""Baseline CLI for Databricks Labs projects."""

import functools
import inspect
import json
import logging
import types
from collections.abc import Callable
from dataclasses import dataclass

from databricks.sdk import AccountClient, WorkspaceClient
from databricks.sdk.config import with_user_agent_extra

from databricks.labs.blueprint.entrypoint import get_logger, run_main
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.blueprint.wheels import ProductInfo


@dataclass
class Command:
    name: str
    description: str
    fn: Callable[..., None]
    is_account: bool = False
    is_unauthenticated: bool = False

    def needs_workspace_client(self):
        if self.is_unauthenticated:
            return False
        if self.is_account:
            return False
        return True

    def prompts_argument_name(self) -> str | None:
        sig = inspect.signature(self.fn)
        for param in sig.parameters.values():
            if param.annotation is Prompts:
                return param.name
        return None

    def get_argument_type(self, argument_name: str) -> str | None:
        sig = inspect.signature(self.fn)
        if argument_name not in sig.parameters:
            return None
        annotation = sig.parameters[argument_name].annotation
        if isinstance(annotation, types.UnionType):
            return str(annotation)
        return annotation.__name__


class App:
    def __init__(self, __file: str):
        self._mapping: dict[str, Command] = {}
        self._logger = get_logger(__file)
        self._product_info = ProductInfo(__file)

    def command(self, fn=None, is_account: bool = False, is_unauthenticated: bool = False):
        """Decorator to register a function as a command."""

        def register(func):
            command_name = func.__name__.replace("_", "-")
            if not func.__doc__:
                raise SyntaxError(f"{func.__name__} must have some doc comment")
            self._mapping[command_name] = Command(
                name=command_name,
                description=func.__doc__,
                fn=func,
                is_account=is_account,
                is_unauthenticated=is_unauthenticated,
            )
            return func

        if fn is None:
            return functools.partial(register)
        register(fn)
        return fn

    def _route(self, raw):
        """Route the command. This is the entry point for the CLI."""
        payload = json.loads(raw)
        command = payload["command"]
        if command not in self._mapping:
            msg = f"cannot find command: {command}"
            raise KeyError(msg)
        # user agent is set consistently with the Databricks CLI:
        # see https://github.com/databricks/cli/blob/main/cmd/root/user_agent_command.go#L35-L37
        with_user_agent_extra("cmd", command)
        flags = payload["flags"]
        log_level = flags.pop("log_level")
        if log_level == "disabled":
            log_level = "info"
        databricks_logger = logging.getLogger("databricks")
        databricks_logger.setLevel(log_level.upper())
        kwargs = {k.replace("-", "_"): v for k, v in flags.items() if v != ""}
        cmd = self._mapping[command]
        # modify kwargs to match the type of the argument
        for kwarg in list(kwargs.keys()):
            match cmd.get_argument_type(kwarg):
                case "int":
                    kwargs[kwarg] = int(kwargs[kwarg])
                case "bool":
                    kwargs[kwarg] = kwargs[kwarg].lower() == "true"
                case "float":
                    kwargs[kwarg] = float(kwargs[kwarg])
        try:
            if cmd.needs_workspace_client():
                kwargs["w"] = self._workspace_client()
            elif cmd.is_account:
                kwargs["a"] = self._account_client()
            prompts_argument = cmd.prompts_argument_name()
            if prompts_argument:
                kwargs[prompts_argument] = Prompts()
            cmd.fn(**kwargs)
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger = self._logger.getChild(command)
            if log_level.lower() in {"debug", "trace"}:
                logger.error(f"Failed to call {command}", exc_info=err)
            else:
                logger.error(f"{err.__class__.__name__}: {err}")

    def _account_client(self):
        return AccountClient(
            product=self._product_info.product_name(),
            product_version=self._product_info.version(),
        )

    def _workspace_client(self):
        return WorkspaceClient(
            product=self._product_info.product_name(),
            product_version=self._product_info.version(),
        )

    def __call__(self):
        run_main(self._route)



"""Product info and wheel builder."""

import inspect
import logging
import random
import shutil
import string
import subprocess
import sys
import tempfile
import warnings
from collections.abc import Iterable
from contextlib import AbstractContextManager
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import cached_property
from pathlib import Path

from databricks.sdk import WorkspaceClient
from databricks.sdk.mixins.compute import SemVer

from databricks.labs.blueprint.entrypoint import find_project_root
from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState

logger = logging.getLogger(__name__)

IGNORE_DIR_NAMES = {
    ".git",
    ".venv",
    ".databricks",
    ".mypy_cache",
    ".github",
    ".idea",
    ".coverage",
    "htmlcov",
    "__pycache__",
    "tests",
}


class SingleSourceVersionError(NotImplementedError):
    pass


class ProductInfo:
    _version_file_names = ["__about__.py", "__version__.py", "version.py"]

    def __init__(self, __file: str, *, github_org: str = "databrickslabs", product_name: str | None = None):
        self._version_file = self._infer_version_file(Path(__file), self._version_file_names)
        self._product_name = product_name
        self._github_org = github_org

    @classmethod
    def from_class(cls, klass: type) -> "ProductInfo":
        """Create a product info with a class used as a starting point to determine location of the version file."""
        return cls(inspect.getfile(klass))

    @classmethod
    def for_testing(cls, klass: type) -> "ProductInfo":
        """Create a product info for testing purposes with a random product name."""
        return cls(inspect.getfile(klass), product_name=cls._make_random(4))

    def checkout_root(self):
        """Returns the root of the project, where .git folder is located."""
        return find_project_root(self._version_file.as_posix())

    def version_file(self) -> Path:
        """Returns the path to a file, where __version__ variable is defined.

        The path to this package can be thought as a way to determine the other
        assets of your deployed wheel.

        See https://packaging.python.org/guides/single-sourcing-package-version/"""
        return self._version_file

    @cached_property
    def _version(self):
        """Returns current version of the project"""
        if not self.is_git_checkout():
            # normal install, downloaded releases won't have the .git folder
            return self.released_version()
        return self.unreleased_version()

    def version(self):
        return self._version

    def as_semver(self) -> SemVer:
        """Returns the version as SemVer object."""
        return SemVer.parse(self.version())

    def product_name(self) -> str:
        """Returns the product name based on the version file folder name."""
        if self._product_name:
            return self._product_name
        version_file_folder = self._version_file.parent
        return version_file_folder.name.replace("_", "-")

    def released_version(self) -> str:
        """Returns the version from the version file."""
        return self._read_version(self._version_file)

    def is_git_checkout(self) -> bool:
        """Returns True if the project is a git checkout."""
        git_config = self.checkout_root() / ".git" / "config"
        return git_config.exists()

    def is_unreleased_version(self) -> bool:
        """Returns True if we are in the git checkout and the version is unreleased."""
        return "+" in self.version()

    def unreleased_version(self) -> str:
        """Returns the unreleased version based on the `git describe --tags` output."""
        try:
            out = subprocess.run(
                ["git", "describe", "--tags"], stdout=subprocess.PIPE, check=True, cwd=self.checkout_root()
            )  # noqa S607
            git_detached_version = out.stdout.decode("utf8")
            return self._semver_and_pep440(git_detached_version)
        except subprocess.CalledProcessError as err:
            logger.warning(
                "Cannot determine unreleased version. This can be fixed by adding "
                " `git fetch --prune --unshallow` to your CI configuration.",
                exc_info=err,
            )
            return self.released_version()

    def current_installation(self, ws: WorkspaceClient) -> Installation:
        """Returns the current installation of the product."""
        return Installation.current(ws, self.product_name())

    def wheels(self, ws: WorkspaceClient) -> "WheelsV2":
        """Returns the wheel builder."""
        return WheelsV2(self.current_installation(ws), self)

    @staticmethod
    def _make_random(k) -> str:
        """Generate a random string of fixed length"""
        # get a random meaningful word from the system dictionary if it exists
        system_wordlist = Path("/usr/share/dict/words")
        if system_wordlist.exists():
            with system_wordlist.open("r", encoding=sys.getdefaultencoding()) as f:
                at_lest_len = [_.lower() for _ in f.read().splitlines() if len(_) > k]
                first = random.choice(at_lest_len)
                second = random.choice(at_lest_len)
                return f"{first}-{second}"
        charset = string.ascii_uppercase + string.ascii_lowercase + string.digits
        return "".join(random.choices(charset, k=int(k)))

    @staticmethod
    def _semver_and_pep440(git_detached_version: str) -> str:
        """Create a version that is both SemVer and PEP440 compliant."""
        detached_version = SemVer.parse(git_detached_version)
        datestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        # new commits on main branch since the last tag
        new_commits = detached_version.pre_release.split("-")[0] if detached_version.pre_release else None
        # show that it's a version different from the released one in stats
        bump_patch = detached_version.patch + 1
        # create something that is both https://semver.org and https://peps.python.org/pep-0440/
        semver_and_pep0440 = f"{detached_version.major}.{detached_version.minor}.{bump_patch}+{new_commits}{datestamp}"
        # validate the semver
        SemVer.parse(semver_and_pep0440)
        return semver_and_pep0440

    @classmethod
    def _infer_version_file(cls, start: Path, version_file_names: list[str]) -> Path:
        # be aware, that WheelsV2 overwrites this wheel file with unreleased version identifier,
        # if it's a git checkout, so that's why we cannot use __init__.py as version marker file,
        # at least for now.
        for version_file in cls._traverse_up(start, version_file_names):
            try:
                cls._read_version(version_file)
                return version_file
            except SyntaxError:
                continue
        candidates = " or ".join(version_file_names)
        raise SingleSourceVersionError(f"cannot find {candidates} with __version__ variable in the tree of {start}")

    @staticmethod
    def _traverse_up(start: Path, version_file_names: list[str]) -> Iterable[Path]:
        """Traverse up the directory tree and yield the version files."""
        prev_folder = start
        folder = start.parent
        while not folder.samefile(prev_folder):
            for name in version_file_names:
                candidate = folder / name
                if not candidate.exists():
                    continue
                yield candidate
            prev_folder = folder
            folder = folder.parent

    @staticmethod
    def _read_version(version_file: Path) -> str:
        """Read the version from the version file."""
        version_data: dict[str, str] = {}
        with version_file.open("r") as f:
            exec(f.read(), version_data)  # pylint: disable=exec-used
        if "__version__" not in version_data:
            raise SyntaxError("Cannot find __version__")
        return version_data["__version__"]


@dataclass
class Version:
    version: str
    wheel: str
    date: str

    def as_semver(self) -> SemVer:
        return SemVer.parse(self.version)


class WheelsV2(AbstractContextManager):
    """Wheel builder"""

    __version: str | None = None

    def __init__(self, installation: Installation, product_info: ProductInfo, *, verbose: bool = False):
        self._installation = installation
        self._product_info = product_info
        self._verbose = verbose

    def upload_to_dbfs(self) -> str:
        """Uploads the wheel to DBFS location of installation and returns the remote path."""
        with self._local_wheel.open("rb") as f:
            return self._installation.upload_dbfs(f"wheels/{self._local_wheel.name}", f)

    def upload_to_wsfs(self) -> str:
        """Uploads the wheel to WSFS location of installation and returns the remote path."""
        with self._local_wheel.open("rb") as f:
            remote_wheel = self._installation.upload(f"wheels/{self._local_wheel.name}", f.read())
            self._installation.save(Version(self._current_version, remote_wheel, self._now_iso()))
            return remote_wheel

    def upload_wheel_dependencies(self, prefixes: list[str]) -> list[str]:
        """Uploads the wheel dependencies to WSFS location of installation and returns the remote paths.
        :param prefixes : A list of prefixes to match against the wheel names. If a prefix matches, the wheel is uploaded.
        """
        remote_paths = []
        for wheel in self._build_wheel(self._tmp_dir.name, verbose=self._verbose, no_deps=False, dirs_exist_ok=True):
            if not wheel.name.endswith("-none-any.whl"):
                continue
            # main wheel is uploaded with upload_to_wsfs() method.
            if wheel.name == self._local_wheel.name:
                continue
            for prefix in prefixes:
                if not wheel.name.startswith(prefix):
                    continue
                remote_wheel = self._installation.upload(f"wheels/{wheel.name}", wheel.read_bytes())
                remote_paths.append(remote_wheel)
        return remote_paths

    @cached_property
    def _current_version(self):
        # addresses double-uploaded bug for unreleased versions uploaded to airgapped workspaces
        return self._product_info.version()

    @staticmethod
    def _now_iso():
        """Returns the current time in ISO format."""
        return datetime.now(timezone.utc).isoformat()

    def __enter__(self) -> "WheelsV2":
        """Builds the wheel and returns the instance. Use it as a context manager."""
        self._tmp_dir = tempfile.TemporaryDirectory()
        self._local_wheel = next(self._build_wheel(self._tmp_dir.name, verbose=self._verbose, no_deps=True))
        return self

    def __exit__(self, __exc_type, __exc_value, __traceback):
        """Cleans up the temporary directory. Use it as a context manager."""
        self._tmp_dir.cleanup()

    def _build_wheel(self, tmp_dir: str, *, verbose: bool = False, no_deps: bool = True, dirs_exist_ok: bool = False):
        """Helper to build the wheel package

        :param tmp_dir: str:
        :param *:
        :param verbose: bool:  (Default value = False)
        :param no_deps: bool:  (Default value = True)
        :param dirs_exist_ok: bool:  (Default value = False)
        """
        stdout = subprocess.STDOUT
        stderr = subprocess.STDOUT
        if not verbose:
            stdout = subprocess.DEVNULL
            stderr = subprocess.DEVNULL
        checkout_root = self._product_info.checkout_root()
        if self._product_info.is_git_checkout() and self._product_info.is_unreleased_version():
            # working copy becomes project root for building a wheel
            checkout_root = self._copy_root_to(tmp_dir, dirs_exist_ok)
            # and override the version file
            self._override_version_to_unreleased(checkout_root)
        args = [sys.executable, "-m", "pip", "wheel", "--wheel-dir", tmp_dir, checkout_root.as_posix()]
        logger.debug(f"Building wheel for {checkout_root} in {tmp_dir}")
        if no_deps:
            args.append("--no-deps")
        subprocess.run(
            args,
            check=True,
            stdout=stdout,
            stderr=stderr,
        )
        return Path(tmp_dir).glob("*.whl")

    def _override_version_to_unreleased(self, tmp_dir_path: Path):
        """Overrides the version file to unreleased version."""
        checkout_root = self._product_info.checkout_root()
        relative_version_file = self._product_info.version_file().relative_to(checkout_root)
        version_file = tmp_dir_path / relative_version_file
        with version_file.open("w") as f:
            f.write(f'__version__ = "{self._current_version}"')

    def _copy_root_to(self, tmp_dir: str | Path, dirs_exist_ok: bool = False):
        """Copies the root to a temporary directory."""
        checkout_root = self._product_info.checkout_root()
        tmp_dir_path = Path(tmp_dir) / "working-copy"

        # copy everything to a temporary directory
        def copy_ignore(_, names: list[str]):
            # callable(src, names) -> ignored_names
            ignored_names = []
            for name in names:
                if name not in IGNORE_DIR_NAMES:
                    continue
                ignored_names.append(name)
            return ignored_names

        shutil.copytree(checkout_root, tmp_dir_path, ignore=copy_ignore, dirs_exist_ok=dirs_exist_ok)
        return tmp_dir_path


class Wheels(WheelsV2):
    """Wheel builder"""

    def __init__(
        self,
        ws: WorkspaceClient,
        install_state: InstallState,
        product_info: ProductInfo,
        *,
        verbose: bool = False,
    ):
        warnings.warn("Wheels is deprecated, use WheelsV2 instead", DeprecationWarning)
        installation = Installation(ws, product_info.product_name(), install_folder=install_state.install_folder())
        super().__init__(installation, product_info, verbose=verbose)
